# Binance
BINANCE_SPOT_BASE_URL = "https://api.binance.com"
BINANCE_FUT_BASE_URL = "https://fapi.binance.com"
BINANCE_SPOT_TEST_BASE_URL = "https://testnet.binance.vision"
BINANCE_FUT_TEST_BASE_URL = "https://testnet.binancefuture.com"

# Phemex
PHEMEX_FUT_BASE_URL = "https://vapi.phemex.com"
PHEMEX_FUT_TEST_BASE_URL = "https://testnet-api.phemex.com"

# Kucoin
KUCOIN_FUT_BASE_URL = "https://api-futures.kucoin.com"
KUCOIN_FUT_TEST_BASE_URL = "https://api-sandbox-futures.kucoin.com"

# Bybit
BYBIT_FUT_BASE_URL = "https://api.bybit.com"
BYBIT_FUT_TEST_BASE_URL = "https://api-testnet.bybit.com"

# FTX
FTX_FUT_BASE_URL = "https://ftx.com/api"

# FTX-US
FTXUS_FUT_BASE_URL = "https://ftx.us/api"
